structure FixContext (* :> FIX_CONTEXT *) =
struct

    (* Nonfix bindings just shadow Infix bindings -- contexts only grow *)
    datatype fixbinding =
        Nonfix
      | Infix of int * Parsing.associativity

    type context = (string * fixbinding) list

    fun isInfix ctx id =
        case ListUtil.Alist.find (op= : string * string -> bool) ctx id of
            SOME (Infix _) => true
          | _ => false

    fun getInfix ctx id =
        case ListUtil.Alist.find (op= : string * string -> bool) ctx id of
            SOME (Infix precass) => SOME (id, precass)
          | _ => NONE

    fun addInfix ctx id precass = (id, Infix precass) :: ctx
    fun delInfix ctx id         = (id, Nonfix) :: ctx

    exception InvalidArg of context * context
    fun diff ctx1 ctx2 =
        let val l1 = length ctx1
            val l2 = length ctx2
            val diff = l1 - l2
        in
            if diff < 0 then
                raise InvalidArg (ctx1, ctx2)
            else
                List.take (ctx1, diff)
        end

    fun join ctx1 ctx2 = ctx1 @ ctx2

    (* mostly borrowed from Andreas Rossberg's HaMLet's basis/infix.sml *)
    val initial =
        [
         (":=", Infix (3, Parsing.Non)),
         ("o", Infix (3, Parsing.Left)),

         ("=", Infix (4, Parsing.Non)),
         ("<>", Infix (4, Parsing.Non)),
         (">", Infix (4, Parsing.Non)),
         (">=", Infix (4, Parsing.Non)),
         ("<", Infix (4, Parsing.Non)),
         ("<=", Infix (4, Parsing.Non)),

         ("::", Infix (5, Parsing.Right)),
         ("@", Infix (5, Parsing.Right)),

         ("+", Infix (6, Parsing.Left)),
         ("-", Infix (6, Parsing.Left)),
         ("^", Infix (6, Parsing.Left)),

         ("*", Infix (7, Parsing.Left)),
         ("/", Infix (7, Parsing.Left)),
         ("div", Infix (7, Parsing.Left)),
         ("mod", Infix (7, Parsing.Left))
        ]

end
