structure Print :> PRINT =
  struct
    open Fomega

    exception Print of string

    val $ = Layout.str
    val % = Layout.mayAlign

    fun var x = $ (Variable.toString x)

    fun atomic_kind_to_layout knd =
        case knd of
            KndType => $"Type"
          | _ => Layout.paren (kind_to_layout knd)

    and arrow_kind_to_layout knd =
        case knd of
            KndArrow (k1, k2) =>
                let fun collect k =
                        case k of
                            KndArrow (k1, k2) =>
                                let val (ks, last) = collect k2
                                in
                                    (k1 :: ks, last)
                                end
                          | _ => ([], k)
                    
                    val (k2s, last) = collect k2
                    val k2ls = map (fn k =>
                                        Layout.seq [$"=> ",
                                                    atomic_kind_to_layout k])
                                   k2s
                    val lastl = Layout.seq [$"=> ", kind_to_layout last]
                in
                    %(atomic_kind_to_layout k1 :: k2ls @ [lastl])
                end
          | _ => atomic_kind_to_layout knd

    and kind_to_layout knd = arrow_kind_to_layout knd


    fun atomic_con_to_layout con =
        case con of
            ConVar cx => var cx
          | _ => Layout.paren (con_to_layout con)

    and app_con_to_layout con =
        case con of
            ConApp (c1, c2) =>
                %[app_con_to_layout c1,
                  atomic_con_to_layout c2]
          | _ => atomic_con_to_layout con

    and arrow_con_to_layout con =
        case con of
            TyArrow (c1, c2) =>
                let fun collect c =
                        case c of 
                            TyArrow (c1, c2) =>
                                let val (cs, last) = collect c2
                                in
                                    (c1 :: cs, last)
                                end
                          | _ => ([], c)
                    
                    val (c2s, last) = collect c2
                    val c2ls = map (fn c =>
                                        Layout.seq [$"-> ",
                                                    app_con_to_layout c])
                                   c2s
                    val lastl = Layout.seq [$"-> ", con_to_layout last]
                in
                    %(app_con_to_layout c1 :: c2ls @ [lastl])
                end
                (*
                %[app_con_to_layout c1,
                  Layout.seq [$"-> ", con_to_layout c2]]
                *)
          | _ => app_con_to_layout con

    and con_to_layout con =
        case con of
            TyForall (cx, k, c) =>
              %[Layout.seq [$"All ", var cx, $"::", kind_to_layout k, $"."],
                con_to_layout c]
          | ConAbs (cx, k, c) =>
              %[Layout.seq [$"lambda ", var cx, $"::", kind_to_layout k, $"."],
                con_to_layout c]
          | _ => arrow_con_to_layout con


    fun atomic_exp_to_layout exp =
        case exp of
            ExVar ex => var ex
          | _ => Layout.paren (exp_to_layout exp)

    and app_exp_to_layout exp =
        case exp of
            ExApp (e1, e2) =>
                %[app_exp_to_layout e1,
                  atomic_exp_to_layout e2]
          | ExConApp (e, c) =>
                %[app_exp_to_layout e,
                  Layout.seq [$"[", con_to_layout c, $"]"]]
          | _ => atomic_exp_to_layout exp

    and exp_to_layout exp =
        case exp of
            ExConAbs (cx, k, e) =>
                %[Layout.seq [$"Lambda ", var cx, $"::",
                              kind_to_layout k, $"."],
                  exp_to_layout e]
          | ExAbs (x, c, e) =>
                %[Layout.seq [$"lambda ", var x, $":",
                              arrow_con_to_layout c, $"."],
                  exp_to_layout e]
          | _ => app_exp_to_layout exp


    val kind_to_string = fn k => Layout.tostring (kind_to_layout k)
    val con_to_string = fn c => Layout.tostring (con_to_layout c)
    val exp_to_string = fn e => Layout.tostring (exp_to_layout e)

    fun print_kind k = print (kind_to_string k)
    fun print_con c = print (con_to_string c)
    fun print_exp e = print (exp_to_string e)
  end
