structure Top :> TOP =
  struct
    exception Unimplemented

    val ctxempty = Typecheck.Ctx.empty

    fun check_exp exp =
        let val ty = Typecheck.check_exp ctxempty exp
            val layout =
                Layout.align
                    [Layout.seq [Layout.str "|- ", Print.exp_to_layout exp],
                     Layout.seq [Layout.str " : ", Print.con_to_layout ty]]
        in
            Layout.print (layout, print);
            print "\n"
            (*
            print "|- ";
            Print.print_exp exp;
            print "\n : ";
            Print.print_con ty;
            print "\n\n"
            *)
        end

    fun process_program parse input =
        let val exps = parse input
        in
            List.app check_exp exps
        end
        handle Parser.Parse (pos, string) =>
                print ("Syntax error: at "
                       ^ Pos.toString pos ^ ": " ^ string ^ "\n")
             | Tokenize.Lex (pos, string) =>
                print ("Syntax error: at "
                       ^ Pos.toString pos ^ ": " ^ string ^ "\n")
             | Typecheck.Check string =>
                print ("Type error: " ^ string ^ "\n")
             | Typecheck.Error string =>
                print ("Internal error: " ^ string ^ "\n")

    fun typecheck_string s =
        process_program (Parser.parseString Parser.parseProgram) s
    fun typecheck_file file =
        process_program (Parser.parseFile Parser.parseProgram) file

    fun equiv_strings s1 s2 =
        let exception Failed
            fun parse s inf =
                Parser.parseString Parser.parseCon s
                handle
                  Parser.Parse (pos, string) =>
                    (print ("Syntax error in " ^ inf ^ ": at "
                           ^ Pos.toString pos ^ ": " ^ string ^ "\n");
                     raise Failed)
                | Tokenize.Lex (pos, string) =>
                    (print ("Syntax error in " ^ inf ^ ": at "
                           ^ Pos.toString pos ^ ": " ^ string ^ "\n");
                     raise Failed)
            fun equiv c1 c2 =
                let val k1 = Typecheck.check_con ctxempty c1
                    val k2 = Typecheck.check_con ctxempty c2
                in
                    if k1 = k2 then
                        Typecheck.equiv_con ctxempty c1 c2 k1
                    else
                        (print ("Error: cons have different kinds\n");
                         raise Failed)
                end
                handle Typecheck.Check string =>
                            (print ("Type error: " ^ string ^ "\n");
                             raise Failed)
                     | Typecheck.Error string =>
                            (print ("Internal error: " ^ string ^ "\n");
                             raise Failed)
            val con1 = parse s1 "first string"
            val con2 = parse s2 "second string"
        in
            equiv con1 con2
        end
        handle Failed => false
  end
